# -*- coding: utf-8 -*-
#
# Generic Shell
# Copyright (C) 2010 Sébastien Dumont <sdumont@masteri2l.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

## \file shell.py
# \author Sébastien Dumont
# \version 1.0
# \date 01/02/2010
# \brief définit les classes de gestion du shell

# Import from the Standard Library
from sys import exit, stdout

# Import from here
from exceptions import CommandError, CommandNotFoundError, TooFewArgError
from exceptions import TooManyArgError

##############################################################################
# Basic command
##############################################################################

##\class ShellCommand
#\brief gère les commandes shell et leurs arguments
class ShellCommand(object):
    """ A basic command.
    """
    MIN_ARG = 0
    MAX_ARG = 0

    ##\brief appelle la méthode validate_arguments
    #\param args contient les paramètres de la commande
    def process(self, args):
        """ Executed method when a command is lauched by the user.
        """
        self.validate_arguments(args)


    ##\brief vérifie si le nombre de paramètres est correct
    #\param args contient les paramètres de la commande
    def validate_arguments(self, args):
        """ A validating method which raise an exception in case of error.
        """
        if len(args) < self.MIN_ARG:
            raise TooFewArgError
        elif len(args) > self.MAX_ARG:
            raise TooManyArgError


    ##\brief renvoi le docstring de la classe
    def help(self):
        """ The help method. It read the docstring of the current class.
        """
        return self.__doc__.strip()


##\class HelpCommand
#\brief liste les commandes possibles et l'aide associée
class HelpCommand(ShellCommand):
    """ List available commands
    """
    MIN_ARG = 0
    MAX_ARG = 1

    ##\brief contructeur
    #\param shell
    def __init__(self, shell):
        ShellCommand.__init__(self)
        self.command_list = shell.command_list

    ##\brief Affiche la liste des commandes ainsi que l'aide de chacune
    #\param args contient les paramêtres de la commande
    def process(self, args):
        ShellCommand.process(self, args)
        if args:
            commands = args
        else:
            commands = self.command_list.keys()

        for command_name in commands:
            command_description = self.command_list[command_name].help()
            print("%s %s" % (command_name, command_description))


    ##\brief Vérifie si la commande existe
    #\param args contient les paramêtres de la commande
    def validate_arguments(self, args):
        ShellCommand.validate_arguments(self, args)
        if args and args[0] not in self.command_list:
            raise CommandNotFoundError



##\class ExitCommand
#\brief quitte le programme
class ExitCommand(ShellCommand):
    """ Exit the program
    """
    MIN_ARG = 0
    MAX_ARG = 0

    ##\brief Ferme le programme et renvoie 0 en code de retour
    def process(self, args):
        exit(0)



##############################################################################
# Basic shell
##############################################################################
##\class Shell
#\brief gère l'interpréteur de commande
class Shell(object):

    ##\brief contructeur
    #\param command_list Liste des commandes
    #\param start_string Prompt du shell
    def __init__(self, command_list = None, start_string = '>> '):
        # Set basic commands
        self.command_list = {}
        self.command_list['help'] = HelpCommand(self)
        self.command_list['quit'] = ExitCommand()
        if command_list is not None:
            self.command_list.update(command_list)
        self.start_string = start_string

    ##\brief affiche l'interpréteur de commande
    def shell_loop(self):
        while True:
            # get stdin
            stdout.write(self.start_string)
            try:
                line = raw_input()
            except EOFError:
                break
            line = line.split()
            if line:
                command, args = line[0], line[1:]
                try:
                    self.run_command(command, args)
                except CommandError, msg:
                    print msg


    ##\brief Lance la commande si elle est bien présente dans la liste des commandes
    #\param command Commande à lancer
    #\param args Arguments de la commande
    def run_command(self, command, args):
        # TODO remove debug
        # print "DEBUG command: '%s' args: '%s'" % (command, args)
        if command in self.command_list:
            self.command_list[command].process(args)
        else:
            raise CommandNotFoundError



