/*----------------------------------------------------------------------*/
/* Fichier : balsemshm.c                                                */
/* Auteur  : D. Duvivier - LIL/ULCO                                     */
/* Date    : 14/10/07                                                   */
/* Objectif: TP N3 M2/I2L Systme Rseaux Scurit                     */
/*           Test d'utilisation d'un tableau de quatre smaphores entre */
/*           quatre processus partageant un segment de mmoire utilis  */
/*           comme "boite aux lettres" (bal) unidirectionnelle          */
/*           (c'est--dire que seuls les producteurs peuvent y dposer  */
/*           des messages  destination du consommateur).               */
/*           Dans cet exemple, il y a deux producteurs, un consommateur */
/*           et un observateur/consultant qui permet de "superviser" ou */
/*           de visualiser rgulirement le nombre de messages dans la  */
/*           boite aux lettres.                                         */
/*           Les quatre smaphores doivent assurer une lecture (et une  */
/*           seule) de chaque message crit. Ils assurent galement de  */
/*           bloquer le lecteur tant qu'il n'y a rien  lire. Ils       */
/*           assurent galement le blocage des producteurs lorsque la   */
/*           boite aux lettres est pleine. Et enfin ils assurent        */
/*           l'accs exclusif (en mutex) des processus dans le moniteur */
/*           Le tableau de quatre smaphores permet de raliser un      */
/*           "pseudo-moniteur" autour de la file circulaire stocke     */
/*           dans le segment de mmoire partag et jouant le            */
/*           rle d'une BAL. Les smaphores sont utiliss comme suit :  */
/*           - semaphore 0 : accs en mutex au moniteur de tous les     */
/*             processus                                                */
/*           - semaphore 1 : accs de NB_MESSAGES processus producteurs */
/*             au plus (en fonction du nombre de messages dj en BAL)  */
/*           - semaphore 2 : accs au moniteur du processus             */
/*             consommateur                                             */
/*           - semaphore 3 : accs au moniteur du processus consultant  */
/*           Ce programme lance les producteurs de donnes et le        */
/*           consommateur de donnes (via un ensemble de fork) et il se */
/*           met en attente afin de supprimer les smaphores et le      */
/*           segment de mmoire lorsque tout est termin.               */
/*           J'ajoute un processus "consultant" qui vient consulter     */
/*           priodiquement le nombre de messages en bal (pour faire de */
/*           la supervision par exemple dans un cal "rel").            */
/*           Cette version utilise une variable "NP" situe dans le     */
/*           moniteur bal pour calculer le nombre de producteurs        */
/*           autoriss  dposer en bal (c'est  dire dbloqus de sem1 */
/*           mais toujours bloqu sur sem0), sinon un bug se produit    */
/*           lorsque la taille de la bal est < au nombre de messages    */
/*           autoriss  tre dposs en bal.                           */
/*           Cette version utilise une variable "AncienneTaille"        */
/*           situe dans le moniteur pour dterminer le type de         */
/*           processus (producteur, consommateur ou consultant) sortant */
/*           sortant du moniteur :                                      */
/*           - si AncienneTaille > Taille alors c'est un Consommateur   */
/*           - si AncienneTaille < Taille alors c'est un Producteur     */
/*           - si AncienneTaille = Taille alors c'est un Consultant     */
/*           Bien videmment il existe d'autres moyens plus "gnriques"*/
/*           de procder, l'intrt ici est de montrer la manipulation  */
/*           d'une variable au sein du moniteur via les primitives      */
/*           bal_entrer() et bal_sortir()                               */
/*----------------------------------------------------------------------*/
/* Remarque : les appels "fflush(stdout)" ne servent  rien si ce n'est */
/* pour s'assurer que les messages soient bien affichs immdiatement   */
/* en vitant les dlais ds  la buffrisation des e/s, ceci est       */
/* particulirement utile pour localiser les erreurs lors d'un plantage */
/* gnral (du genre "erreur de segmentation) sans dbugger.            */
/*----------------------------------------------------------------------*/
/* Remarque : cette version n'accde plus  semid et shmid comme des    */
/* variables globales (afin de prparer l'utilisation simultane de     */
/* plusieurs tableaux de smaphores pour implmenter plusieurs bal).    */
/* semid est maintenant stock dans la structure "bal" ce qui est plus  */
/* logique/correct/modulaire/rutilisable car le tableau de smaphores  */
/* DOIT faire partie intgrante de la strcture de donnes associe  une*/
/* bal.                                                                 */
/*----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include <signal.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>

         /* La "constante" ci-dessous indique le nombre de smaphores */
         /* dans le tableau de smaphores identifi par semid */
#define  NB_SEMAPHORES 4

         /* Nombre de producteurs */
#define  NB_PRODUCTEURS 2

         /* Longueur (maxi) des messages dposs (multiple de 8) */
#define  LNG_MESSAGES 128

         /* Taille (en octets) du segment mmoire (multiple de 8) */
#define  SEG_SIZE 1408

         /* Nombre maxi de messages calcul  partir */
         /* de la taille du segment mmoire, de la longueur des */
         /* messages, en dcomptant la taille de 6 entiers */
         /* (semid, NP, AncienneTaille, tete, queue et taille) utiliss*/
         /* pour grer les msg en bal */
#define  NB_MESSAGES ((int)(((int)(SEG_SIZE - (6*sizeof(int))))/((int)LNG_MESSAGES)))

         /* Mode dbuggage (affiche des messages d'tats des process) : */
         /* Remplacez "undef" par "define" en ci-dessous pour passer en */
         /* mode dbuggage (affiche "plein de messages") et inversement */
#undef   DD_DEBUG

         /* Type d'un message : buffer/tableau de LNG_MESSAGES caractres*/
typedef  char   TDD_message[LNG_MESSAGES];

         /* Type d'une bote aux lettres : c'est une file circulaire de */
         /* messages implmente dans un tableau de taille fixe */
         /* REM : il est possible d'utiliser une "queue/file" systme */
         /* (ou un tube +/- nomm) */
typedef  struct SDD_bal {
  int semid;    /* Indentificateur du tableau de smaphores partag */
                /* permettant de simuler un moniteur "bal" */
  TDD_message   msg[NB_MESSAGES]; /* Tableau de messages */
  int           tete, queue;      /* Tte et queue de file */
  int           taille;           /* Nombre de message en file */
  int           AncienneTaille;   /* Taille lors de l'entre en moniteur*/
  int           NP;               /* Nombre de producteurs autoriss  */
                                  /* dposer un message en bal */
}        TDD_bal;

/* Affiche un message d'information */
#define pidmsg2(a, b)   \
        printf("%-12s:pid=%d:%s\n", a, getpid(), b);\
        fflush(stdout);

#define pidmsg3(a, b, c)        \
        printf("%-12s:pid=%d:%s%d)\n", a, getpid(), b, c);\
        fflush(stdout);

/* Affiche un message d'information si DD_DEBUG est dfini */
/* sinon cette macro ne fait rien (!) */
#ifdef DD_DEBUG
#define pidmsgdebug2(a, b)      \
        printf("%-12s:pid=%d:%s\n", a, getpid(), b);\
        fflush(stdout);
#define pidmsgdebug3(a, b, c)   \
        printf("%-12s:pid=%d:%s%d)\n", a, getpid(), b, c);\
        fflush(stdout);
#else
#define pidmsgdebug2(a, b)
#define pidmsgdebug3(a, b, c)
#endif

/************************************************************************/
/* Variables globales -> IL N'Y EN N'A PLUS !!!                         */
/* (il faut de manire gnrale limiter leur nombre au strict ncessaire*/
/************************************************************************/

/* Acquisition d'un segment de mmoire partage (LM49) */
/* ATTENTION : le descripteur associ est retourn via la variable */
/* "shmid" */
int shm_init(int *shmid, char *filename, char project_id, int size) {
  pidmsgdebug2("Processus","shm_init()");
  if (*shmid == -1) {
    /* Premier appel -> acquisition du segment */
    key_t k;
    *shmid = shmget(k = ftok(filename, project_id), size, 0600 | IPC_CREAT);
    if (*shmid <= -1) {
      /* Erreur lors de l'acquisition */
      perror("Impossible d'obtenir un segment mmoire");
      exit(EXIT_FAILURE);
    }
    else {
      /* OK : on a un segment ! */
      pidmsgdebug3("Processus","shm_init():Segment mmoire allou correctement (shmid=", *shmid);
      return (1); /* Signale nouveau segment  l'appelant */
    }
  }
  else {
    /* Un segment existe dj -> le rutiliser ! */
#ifdef DD_DEBUG
    pidmsgdebug3("Processus","shm_init():Segment mmoire r-utilis (shmid=", *shmid);
#endif
    return (0); /* Signale r-utilisation d'un segment  l'appelant */
  }
}

/* Suppression du segment mmoire dont l'id est pass via la */
/* variable globale shmid. shmid et rinitialis  -1 par cette procdure */
void shm_destroy(int *shmid) {
  pidmsgdebug2("Processus","shm_destroy()");
  shmctl(*shmid, IPC_RMID, NULL);
  *shmid = -1; /* Scurit : indique segment non allou */
}

/* Acquisition de NB_SEMAPHORES smaphore(s) (LM36) */
/* ATTENTION : le descripteur associ est retourn dans la variable */
/* "semid" */
int sem_init(int *semid, char *filename, char project_id) {
  key_t k;
  pidmsgdebug2("Processus","sem_init()");
  if (*semid == -1) {
    /* Accder au tableau de smaphores s'il existe dj */
    *semid = semget(k = ftok(filename, project_id), NB_SEMAPHORES, 0600);
    if (*semid == -1) {
      /* Premier appel -> acquisition du tableau de smaphores */
      *semid = semget(k = ftok(filename, project_id), NB_SEMAPHORES, 0600 | IPC_CREAT);
      if (*semid <= -1) {
        /* Erreur lors de l'acquisition */
        perror("sem_init():Impossible d'obtenir des smaphores");
        exit(EXIT_FAILURE);
      }
      else {
        /* OK : on a des smaphores ! */
        pidmsgdebug3("Processus","sem_init():Tableau de smaphores allou correctement (semid=", *semid);
        return (1); /* Signale nouveau tableau de smaphores  l'appelant */
      }
    }
    else {
      /* Le tableau de smaphores existe dj et c'est le premier appel  sem_init -> le rutiliser ! */
      pidmsgdebug3("Processus","sem_init() Tableau de smaphores r-utilis (semid=", *semid);
      return (0); /* Signale r-utilisation des smaphores  l'appelant */
    }
  }
  else {
    /* Le tableau de smaphores existe dj et ce n'est pas le premier appel  sem_init -> le rutiliser ! */
    pidmsgdebug3("Processus","Tableau de smaphores r-utilis (semid=", *semid);
    return (-1); /* Signale r-utilisation des smaphores  l'appelant */
                 /* et appels multiples de sem_init */
  }
}

/* Suppression du tableau de smaphores dont l'id est pass via la*/
/* variable semid. semid est rinitialis  -1 en fin de procdure */
/* pour indiquer que le tableau de smaphores n'est plus allou */
void sem_destroy(int *semid) {
  pidmsgdebug2("Processus","sem_destroy()");
  semctl(*semid, 0, IPC_RMID, 0);
  *semid = -1;
}

/* Fixe la valeur initiale du smaphore numro "sem" relatif au tableau de */
/* smaphores identifi par "semid" */
void sem_setvalue(int semid, int sem, int val) {
#if defined(__GNU_LIBRARY__) && !defined(_SEM_SEMUN_UNDEFINED)
  /* l'union semun est dfinie par l'inclusion de <sys/sem.h> */
  union  semun arg;
#else
  /* d'aprs X/OPEN il faut la dfinir nous-mmes */
  union semun {
        int val;                  /* valeur pour SETVAL */
        struct semid_ds *buf;     /* buffer pour IPC_STAT, IPC_SET */
        unsigned short *array;    /* table  pour GETALL, SETALL */
                                  /* Spcificit Linux : */
        struct seminfo *__buf;    /* buffer pour IPC_INFO */
  } arg;
#endif
  pidmsgdebug3("Processus","sem_setvalue(sem",sem);
  arg.val = val; /* Fixe la valeur initiale du smaphore */
  if (semctl(semid, sem, SETVAL, arg) <= -1) {
    perror("Impossible de fixer la valeur initiale du smaphore");
    exit(EXIT_FAILURE);
  }
}

/* Lit "la valeur" initiale du smaphore numro "sem" relatif au tableau de */
/* smaphores identifi par "semid" */
int sem_getvalue(int semid, int sem) {
  int val; /* "Valeur" du smaphore */
  pidmsgdebug3("Processus","sem_getvalue(sem",sem);
  if ((val = semctl(semid, sem, GETVAL)) <= -1) {
    perror("Impossible de lire la valeur du smaphore");
    exit(EXIT_FAILURE);
  }
  return (val); /* Retourne la valeur du smaphore */
}

/* Affiche "les valeurs" des smaphores du tableau semid */
void sem_printvalues(int semid) {
  int i; /* Indice de boucle */

  pidmsgdebug2("Processus","sem_printvalues()");
  for (i=0; i<NB_SEMAPHORES; i++) {
    printf("Processus   :pid=%d:sem_printvalues():valeur smaphore %d = %d\n", getpid(), i, sem_getvalue(semid, i));
    fflush(stdout);
  }
}

/* P(smaphore[sem]) */
int sem_askresource (int semid, int sem) {
  int r;
  struct sembuf s;
  pidmsgdebug3("Processus","sem_askresource(sem", sem);
  s.sem_num = sem;
  s.sem_op = -1;
  s.sem_flg = 0;
  if ((r = semop(semid, &s, 1)) <= -1) {
    perror("sem_askresource():Impossible de rserver la ressource");
    exit(EXIT_FAILURE);
  }
  return r;
}

/* V(smaphore[sem]) */
int sem_releaseresource (int semid, int sem) {
  int r;
  struct sembuf s;
  pidmsgdebug3("Processus","sem_release(sem", sem);
  s.sem_num = sem;
  s.sem_op = 1;
  s.sem_flg = 0;
  if ((r = semop(semid, &s, 1)) <= -1) {
    perror("sem_releaseresource():Impossible de librer la ressource");
    exit(EXIT_FAILURE);
  }
  return r;
}

/************************************************************************/
/* Gestion du "pseudo-moniteur" bal                                     */
/************************************************************************/

/* Cre la bote aux lettres (les smaphores et le segment partag) */
/* nomexe est le nom du fichier contenant le programme excutable */
/* (typiquement argv[0]) */
void bal_create(TDD_bal **bal, int *shmid, char *nomexe) {
  pidmsg2("Processus","bal_create()");
  /* On utilise le nom de l'excutable et le numro de projet (ici "1")*/
  /* pour calculer une clef via ftok et obtenir un descripteur de */
  /* segment mmoire unique */
  if (!shm_init(shmid, nomexe, 1, SEG_SIZE)) {
    pidmsg2("Processus","bal_create():segment partag dj attach ?\n");
    exit(EXIT_FAILURE);
  }
  pidmsgdebug3("Processus","bal_create():Attache le segment partag au segment de donnes (shmid=", *shmid);
  *bal = shmat(*shmid, NULL, 0);

  (*bal)->queue = (*bal)->tete = -1; /* Indices hors intervalle valide ! */
  (*bal)->NP = (*bal)->AncienneTaille = (*bal)->taille = 0;
  (*bal)->semid = -1; /* NECESSAIRE pour indiquer smaphores non allous (!) */

  /* On utilise le nom de l'excutable et le numro de projet (ici "1")*/
  /* pour calculer une clef via ftok et obtenir un descripteur de */
  /* tableau de smaphores unique */
  if (!sem_init(&((*bal)->semid), nomexe, 1)) {
    pidmsg2("Processus","bal_create():Le tableau de smaphores existe dj\n");
    /* inutile de faire exit(EXIT_FAILURE) ici ce n'est pas une erreur*/
  }

  /* On fixe la valeur initiale des smaphores  0 de manire  bloquer */
  /* tous les processus tant que la bal n'est pas initialise par */
  /* bal_init() */
  sem_setvalue((*bal)->semid, 0, 0); /*Smaphore d'entre en moniteur (bloqu)*/
  sem_setvalue((*bal)->semid, 1, 0); /*Smaphore pour dpos (bloqu) */
  sem_setvalue((*bal)->semid, 2, 0); /*Smaphore pour retrait (bloqu) */
  sem_setvalue((*bal)->semid, 3, 0); /*Smaphore pour nombremsg (bloqu) */
}

/* Initialise la bote aux lettres */
/* ATTENTION les smaphores doivent tous tre bloqus avant l'appel */
/*  cette procdure qui ne gre pas l'exclusion mutuelle  la bal */
/* car elle est la premire procdure  appeler aprs la cration des */
/* smaphores et du segment de mmoire partag. C'est elle qui "dbloque" */
/* les accs  la bal. */
void bal_init(TDD_bal *bal) {
  pidmsg3("Processus","bal_init(nombre max de messages=", NB_MESSAGES);
  bal->queue = bal->tete = -1; /* Indices hors intervalle valide ! */
  bal->NP = bal->AncienneTaille = bal->taille = 0;

  /* La bal est initialise, l'entre dans le moniteur "bal", les dpos*/
  /* et la consultation du nombre de messages sont autoriss : */
  sem_releaseresource(bal->semid, 0); /* V(semaphore[0]) entre moniteur autorise */
  sem_setvalue(bal->semid, 1, NB_MESSAGES); /* V(semaphore[1]) NB_MESSAGES dpos autoriss */
  sem_releaseresource(bal->semid, 3); /* V(semaphore[3]) info nombre msg autoriss */
  sem_printvalues(bal->semid);
}

/* Dtruit la bote aux lettres (les smaphores et le segment mmoire) */
void bal_destroy(TDD_bal *bal, int *shmid) {
  pidmsgdebug3("Processus","bal_destroy():Destruction des smaphores (semid=", bal->semid);
  /* ATTENTION : il faut d'abord dtruire le tableau de smaphores avant*/
  /* le segment de mmoire partag car l'identificateur semid est stock*/
  /* dans la structure "bal" qui est elle mme stocke dans le segment*/
  /* de mmoire partag */
  sem_destroy(&(bal->semid));
  pidmsgdebug3("Processus","bal_destroy():Dtache segment partag du segment de donnes (shmid=", *shmid);
  shmdt(bal);
  pidmsgdebug2("Processus","bal_destroy():Destruction du segment de mmoire partag");
  shm_destroy(shmid);
}

/* Simule l'entre d'un processus dans le "moniteur" bal */
/* REM : passage par adresse pour optimiser le code */
void bal_entrer(TDD_bal *bal) {
  pidmsgdebug2("Processus","bal_entrer()");
  sem_askresource(bal->semid, 0); /* P(semaphore[0]) entre moniteur autorise ? */
  pidmsgdebug2("Processus","bal_entrer(): EN SECTION CRITIQUE DANS LE MONITEUR");
  /* En ce point il y a un seul processus dans le moniteur bal */
  /* car le smaphore 0 se comporte en mutex sur le moniteur pour */
  /* autoriser au plus un seul processus dans le moniteur bal */
  /* S'il y a des "choses  faire" lors de l'entre d'un procesus */
  /* dans le moniteur, c'est ici qu'il faut le faire ! */
  bal->AncienneTaille = bal->taille; /* Taille  l'entre du moniteur */
}

/* Simule la sortie d'un processus du "moniteur" bal */
/* La sortie d'un processus va dbloquer le prochain ou les prochains */
/* REM : passage par adresse pour optimiser le code */
void bal_sortir(TDD_bal *bal) {
  pidmsgdebug2("Processus","bal_sortir()");
  /* En ce point il y a un seul processus dans le moniteur bal */
  /* car le smaphore 0 se comporte en mutex sur le moniteur pour */
  /* autoriser au plus un seul processus dans le moniteur bal */
  /* Et tant que V(semaphore[0]) n'a pas t effectu, le processus */
  /* "courant" reste (seul) dans le moniteur ! */
  /* S'il y a des "choses  faire" lors de la sortie d'un procesus */
  /* du moniteur, c'est ici qu'il faut le faire ! */

  /* Dterminer si le type de processus est un producteur et si tel est*/
  /* le cas dcrmenter le nombre de processus producteurs autoriss  */
  /* dposer un message en bal : */
  if (bal->AncienneTaille < bal->taille) {
    /* Entre l'entre et la sortie du moniteur un processus a incrment*/
    /* la taille de la bal -> un message a t dpos donc le processus*/
    /* sortant est un producteur -> le dcompter */
    if (bal->NP > 0) {
      bal->NP--;
    }
  }

  /* SIMULATION D'UNE REGION CRITIQUE CONDITIONNELLE :
     le/un processus lecteur
     ne peux accder au coeur du moniteur bal que s'il y
      un message  lire -> Ce n'est pas (uniquement) 
     l'entre bal_retirer() qui
     effectue un appel  sem_releaseresource(2) pour autoriser le prochain
     retrait pour fin de retirer mais l'appel  bal_sortir() par n'importe
     quel processus sortant du moniteur ! */

  /* Dfinition des priorits : on dbloque d'abord les processus qui
     retirent les messages (s'il y en a), ils vont s'agglutiner dans la
     file du smaphore 0 pour demander  entrer en moniteur
  */

  /* PAR CONTRE ON NE GERE PAS LES "CONSULTANTS" (nombre msg) ??? */
  /* ca serait plus logique de grer toute la logique de fonctionnement */
  /* de tous les processus dans le moniteur ! */


  /* A FAIRE !!! */

  pidmsgdebug2("Processus","bal_sortir(): SORT DU MONITEUR");
  sem_releaseresource(bal->semid, 0); /* V(semaphore[0]) entre moniteur autorise*/
  pidmsgdebug2("Processus","bal_sortir(): SORTI DU MONITEUR");
}


/* Dpose un message dans la bote aux lettres */
/* ATTENTION l'exclusion mutuelle  la bal est gre via des smaphores */
/* Retourne 0 si tout s'est bien pass; -1 sinon */
/* Rem : passages par adresse pour optimiser le code */
int bal_deposer(TDD_bal *bal, TDD_message *msg) {
  pidmsgdebug2("Processus","bal_deposer()");
  sem_askresource(bal->semid, 1); /* P(semaphore[1]) pour dposer */
  pidmsgdebug2("Processus","bal_deposer(): DEMANDE ACCES AU MONITEUR");
  pidmsgdebug3("Processus","bal_deposer() (valeur smaphore 1 = ", sem_getvalue(bal->semid, 1));
  /* Si l'on arrive en ce point c'est qu'il reste de la place dans la */
  /* file des messages (bal) car P(semaphore[1]) s'est "dbloqu" */

  /* Demander l'entre en moniteur */
  bal_entrer(bal);
  pidmsgdebug2("Processus","bal_deposer():Accs au moniteur");
  usleep(250000); /* 1/4 seconde en section critique */

  if (bal->taille < NB_MESSAGES) {
    bal->taille++;
  }
  else {
    /* Plus de place */

    /* Ici il faudrait sortir du moniteur et se rebloquer en file ?
       aprs avoir libr le moniteur -> LAISSE EN EXERCICE ! */
    pidmsg2("Processus","bal_deposer():Erreur non gre (bal dborde)\n");
    sem_releaseresource(bal->semid, 1); /* V(semaphore[1]) pour fin de dposer SANS DEPOS */
    pidmsgdebug2("Processus","bal_deposer():Sortie du moniteur");
    /* Sortir du moniteur */
    bal_sortir(bal);
    return -1;
  }
  /* Calcul de l'indice (pr-incrment) pour ajout en queue */    
  if (bal->queue < NB_MESSAGES-1) {
    bal->queue++;
  }
  else {
    bal->queue = 0;
  }
  /* Premier ajout ? */
  if (bal->tete == -1) {
    bal->tete = bal->queue;
  }
  /* Copie (systmatique) des LNG_MESSAGES caractres du message */
  /* pour ajouter le message en queue de file */
  memcpy(bal->msg[bal->queue], msg, LNG_MESSAGES);

  /* SURTOUT NE PAS FAIRE CECI : sem_releaseresource(1); */
  /* car il ne faut utiliser V(semaphore[1]) que lorsqu'un message a t */
  /* lu par le rdacteur sinon la boite dborde ! */
  /* Sortir du moniteur */
  pidmsgdebug2("Processus","bal_deposer(): Sortie du moniteur");
  bal_sortir(bal);
  return 0; /* OK */
}

/* Retire un message de la bote aux lettres */
/* ATTENTION l'exclusion mutuelle  la bal est gre via des smaphores */
/* Retourne 0 si tout s'est bien pass; -1 sinon */
/* Rem : passages par adresse ncessaire ici */
int bal_retirer(TDD_bal *bal, TDD_message *msg) {
  pidmsgdebug2("Processus","bal_retirer()");
  sem_askresource(bal->semid, 2); /* P(semaphore[2]) pour retirer */
  pidmsgdebug2("Processus","bal_retirer(): DEMANDE ACCES AU MONITEUR");
  /* Si l'on arrive en ce point c'est qu'il y a au moins un message */
  /* dans la file des messages (bal) car P(semaphore[2]) s'est "dbloqu"*/

  /* Demander l'entre en moniteur */
  bal_entrer(bal);
  pidmsgdebug2("Processus","bal_retirer():Accs au moniteur");
  usleep(333333); /* 1/3 seconde en section critique */

  if (bal->taille > 0) {
    bal->taille--;
  }
  else {
    /* Plus de message */

    /* Ici il faudrait sortir du moniteur et se rebloquer en file ?
       aprs avoir libr le moniteur -> LAISSE EN EXERCICE ! */
    pidmsg2("Processus","bal_retirer():Erreur non gre (bal vide)\n");

    sem_releaseresource(bal->semid, 2); /* V(semaphore[2]) pour fin de retirer SANS MESSAGE LU */
    pidmsgdebug2("Processus","bal_retirer():Sortie du moniteur");
    /* Sortir du moniteur */
    bal_sortir(bal);
    return -1;
  }
  /* Copie (systmatique) des LNG_MESSAGES caractres du message */
  memcpy(msg, bal->msg[bal->tete], LNG_MESSAGES);

  /* Calcul de l'indice (post-incrment) pour retrait en tte */    
  if (bal->tete < NB_MESSAGES-1) {
    bal->tete++;
  }
  else {
    bal->tete = 0;
  }
  /* Bal vide ? */
  /* Ce test n'est pas ncessaire mais il simplifie le code si l'on */
  /* souhaite autoriser l'accs  la tte et  la queue de la file */
  /* depuis des primitives additionnelle (non implmentes ici) */
  if (bal->taille == 0) {
    bal->tete = bal->queue = -1;
  }

  /* SURTOUT NE PAs FAIRE CECI : sem_releaseresource(2); */
  /* car c'est la logique interne du moniteur qui devrait le */
  /* dcider dans le cadre de gestion de priorits par rgion critiques */
  /* conditionnelles */
  pidmsgdebug2("Processus","bal_retirer():Sortie du moniteur");
  /* Sortir du moniteur */
  bal_sortir(bal);
  return 0; /* OK */
}

/* Indique le nombre de messages dans la boite aux lettres */
/* Rem : passages par adresse pour optimiser le code */
int bal_nombremsg(TDD_bal *bal) {
  int taille;
  pidmsgdebug2("Processus","bal_nombremsg()");
  sem_askresource(bal->semid, 3); /* V(semaphore[3]) info nombremsg autoriss */
  pidmsgdebug2("Processus","bal_nombremsg(): DEMANDE ACCES AU MONITEUR");
  /* Demander l'entre en moniteur */
  bal_entrer(bal);
  pidmsgdebug2("Processus","bal_nombremsg():Accs au moniteur");
  usleep(125000); /* 1/8 seconde en section critique */

  taille = bal->taille;

  /* Soit on laisse ici la ligne suivante : */
  /* sem_releaseresource(3); *//* V(semaphore[3]) info nombremsg autoriss */
  /* soit on gre TOUT dans le moniteur (c'est ce que j'ai fait) */
  pidmsgdebug2("Processus","bal_nombremsg():Sortie du moniteur");
  /* Sortir du moniteur */
  bal_sortir(bal);

  /* En ce point je suis hors de la section critique, */
  /* donc je ne dois plus accder  la structure de donnes "bal" */
  /* par consquent il ne faut SURTOUT pas crire return bal->taille; */
  /* mais ceci : */
  return taille;
}

/************************************************************************/
/* PROCESSUS PRODUCTEUR                                                 */
/************************************************************************/
void Producteur(int shmid) {
  int i; /* Indice de boucle */
  TDD_message un_msg;
  TDD_bal *shmaddr = NULL; /* Adresse de "mappage" du segment */
                        /* (en ralit le pointeur est void *) */ 

  pidmsg2("Producteur","Lancement du code du processus");

#ifdef DD_DEBUG
  pidmsgdebug3("Producteur","Attache le segment partag au segment de donnes (shmid=", shmid);
#endif
  fflush(stdout);
  shmaddr = shmat(shmid, NULL, 0);    
  pidmsg2("Producteur","Envoi des messages");
  for (i = 1; i <= 10; i++) {
    /* Dposer un message dans la bal */
    pidmsg3("Producteur","Demande d'criture dans la bote aux lettres (ecriture #", i);

    sprintf(un_msg, "Message %d du processus %d dans la bote aux lettres\n", i, getpid());

    if (bal_deposer(shmaddr, &un_msg) < 0) {
      pidmsg2("Producteur","Bote aux lettres pleine !");
    }
    pidmsgdebug2("Producteur","Hors de la section critique");
    usleep(300000); /* environ 1/3 seconde hors de section critique */
  }

  /* On dtache le segment mmoire de l'espace d'adressage */
  pidmsgdebug3("Producteur","Dtache le segment partag du segment de donnes (shmid=", shmid);

  shmdt(shmaddr);

  pidmsg2("Producteur","Fin du processus !");
}

/************************************************************************/
/* PROCESSUS CONSOMMATEUR                                               */
/************************************************************************/
void Consommateur(int shmid) {
  int i; /* Indice de boucle */
  TDD_message un_msg;
  TDD_bal *shmaddr = NULL; /* Adresse de "mappage" du segment */
                      /* (en ralit le pointeur est void *) */ 

  pidmsg2("Consommateur","Lancement du code du processus");
  pidmsgdebug3("Consommateur","Attache le segment partag au segment de donnes (shmid=", shmid);

  shmaddr = shmat(shmid, NULL, 0);    

  pidmsg2("Consommateur","Initialisation du moniteur bal");
  /* Initialise la bote aux lettres */
  /* Attention, aucun message ne doit tre dpos avant d'avoir initilis */
  /* la bote aux lettres, ceci est assur par le fait d'initialiser */
  /* le smaphore 1 (demande de dpos)  0 */
  bal_init(shmaddr);

  pidmsg2("Consommateur","Lecture des messages");
  for (i = 1; i <= 10*NB_PRODUCTEURS; i++) {
#ifdef DD_DEBUG
    printf("Consommateur:pid=%d:Demande de lecture #%d via segment partag (shmid=%d) et P(sem2)\n", getpid(), i, shmid);
    fflush(stdout);
#endif

    if (bal_retirer(shmaddr, &un_msg) >= 0) {
      /* Un message a t lu correctement */
      printf("Consommateur:pid=%d:Chane lue=%s", getpid(), un_msg);
      fflush(stdout);
    }
    else {
      /* Pas/plus de message  lire */
      pidmsg2("Consommateur","Plus de message  lire !");
    }
    pidmsgdebug2("Consommateur","Hors de la section critique");
    usleep(100000); /* 1/10 secondes hors de la section critique */
  }
  /* On dtache le segment mmoire de l'espace d'adressage */
  pidmsgdebug3("Consommateur","Dtache le segment partag du segment de donnes (shmid=", shmid);
  shmdt(shmaddr);

  pidmsg2("Consommateur","Fin du processus !");
}

/************************************************************************/
/* PROCESSUS "CONSULTANT"                                               */
/************************************************************************/
void Consultant(int shmid) {
  int i; /* Indice de boucle */
  TDD_bal *shmaddr = NULL; /* Adresse de "mappage" du segment */
                      /* (en ralit le pointeur est void *) */ 

  pidmsg2("Consultant","Lancement du code du processus");
  pidmsgdebug3("Consultant","Attache le segment partag au segment de donnes (shmid=", shmid);
  shmaddr = shmat(shmid, NULL, 0);    

  pidmsg2("Consultant","Consultations priodiques du nombre de messages");
  for (i = 1; i <= 12*NB_PRODUCTEURS; i++) {
#ifdef DD_DEBUG
    printf("Consultant  :pid=%d:Demande de lecture #%d via segment partag (shmid=%d) et P(sem3)\n", getpid(), i, shmid);
    fflush(stdout);
#endif

    /* Un message a t lu correctement */
    pidmsg3("Consultant","(Nombre de messages=", bal_nombremsg(shmaddr));

    pidmsgdebug2("Consultant","Hors de la section critique");
    usleep(250000); /* 1/4 secondes hors de section critique */
  }
  /* On dtache le segment mmoire de l'espace d'adressage */
  pidmsgdebug3("Consultant","Dtache le segment partag du segment de donnes (shmid=", shmid);
  shmdt(shmaddr);

  pidmsg2("Consultant","Fin du processus !");
}

/************************************************************************/
/************************************************************************/
/*                       PROGRAMME PRINCIPAL                            */
/************************************************************************/
/************************************************************************/
int main(int argc, char *argv[]) {
  int i; /* Indice de boucle */
  TDD_bal *shmaddr = NULL; /* Adresse de "mappage" du segment */
                        /* (en ralit le pointeur est de type void *) */ 
  int shmid = -1; /* Indentificateur du segment mmoire partage */
                  /* Commun au pre et aux fils */
  int conspid;    /* Pid du processus fils (i.e. le consommateur) */
  int prodpid[NB_PRODUCTEURS]; /* Pid des processus producteurs */
  int superpid;   /* Pid du processus fils (le consultant/superviseur) */

  pidmsg2("ProcessPre","Lancement du processus pre");
  bal_create(&shmaddr, &shmid, argv[0]);
  if ((shmid <= 0) || (shmaddr == NULL)) {
    pidmsg2("ProcessPre","Erreur lors de la cration de la bal !");
    exit(EXIT_FAILURE);
  }

  /* En ce point nous allons lancer le consommateur  partir du */
  /* pre en effectuant un fork() */
  /* ATTENTION : par soucis de concision, tous les codes de retour et */
  /* toutes le prcautions ncessaires  l'criture d'un programme complet*/
  /* et "robuste" n'ont pas t ajoute  ce programme */
  /* -> allez voir dans les pages de man () la documentation des */
  /* fonctions utilises et les prcautions  mettre en oeuvre */
  /* en guise d'exercice ! */

  conspid = fork(); /* On clone le pre pour obtenir le consommateur */
  if (conspid < 0) {
    fprintf(stderr, "ProcessPre :pid=%d:L'erreur '%s' s'est produite lors de l'appel  fork() pour crer le consommateur\n",
              getpid(), strerror(errno));
    exit(EXIT_FAILURE);
  }
  else if (conspid > 0) {
    /* Code du pre */
    /* En ce point nous allons lancer le consultant  partir du */
    /* pre en effectuant un fork() */
    superpid = fork();
    if (superpid < 0) {
      fprintf(stderr, "ProcessPre :pid=%d:L'erreur '%s' s'est produite lors de l'appel  fork() pour crer le consultant\n",
                getpid(), strerror(errno));
      exit(EXIT_FAILURE);
    }
    else if (superpid > 0) {
      /* Code du pre */
      /* Le pre va lancer tous les producteurs comme des fils. */
      /* C'est un peu diffrent de ce qui tait effectu prcdemment : */
      /* Le pre (c'est--dire le processus initialement lanc depuis le */
      /* se comporte comme un serveur : il lance toutes le tches de calcul*/
      /* dans des processus fils (producteurs, consultants et */
      /* consommateur) et se met en attente des vnement de terminaison */
      /* (...) des "n" processus fils en bouclant "n" fois sur wait() */
      pidmsg2("ProcessPre","Lancement des producteurs");
      for (i=0; i<NB_PRODUCTEURS; i++) {
        prodpid[i] = fork(); /* On clone le pre pour obtenir 1 producteur */
        if (prodpid[i] < 0) {
          fprintf(stderr, "ProcessPre :pid=%d:L'erreur '%s' s'est produite lors de l'appel  fork() pour crer un producteur\n",
                getpid(), strerror(errno));
          exit(EXIT_FAILURE);
        }
        else if (prodpid[i] == 0) {
          /* Code du fils : il faut lancer le producteur ... */
          Producteur(shmid);
          /* et ne surtout pas oublier de stopper la boucle pour */
          /* par un "break;" !!! */
          /* sinon cela va lancer des producteurs en cascade !!! */
          i = -1; /* Indique boucle termine par break dans un fils */
          break; /* -> fin de boucle */
        }
        /* else/sinon nous sommes dans le processus pre, dans lequel la */
        /* boucle for/pour doit se poursuivre pour lancer les */
        /* producteurs ...*/

      } /* for */
      /* Suite du code excut par le pre ET par les fils producteurs */
      /* Si i > 0 alors il s'agit du processus pre */
      if (i > 0) {
        /* Il faut attendre la fin des NB_PRODUCTEURS + 2 fils ...*/
        pidmsg2("ProcessPre","Mise en attente du pre");
        for (i=0; i<=NB_PRODUCTEURS+1; i++) {
          /* Ci-dessous il faudrait intercepter les codes de retour et */
          /* vrifier si la sortie de wait est bien due  l'arrt d'un */
          /* processus fils -> A faire en exercice ! */
          wait(NULL);
          pidmsg2("ProcessPre","Un processus fils vient de se terminer...");
        }
        pidmsg2("ProcessPre","Tous les processus fils sont termins");
        /* Il faut faire le mnage ... */
        bal_destroy(shmaddr, &shmid);
      } /* if (i > 0) */
    } /* if (superpid > 0) */
    else {
      /* Code du fils consultant */
      Consultant(shmid);
    }
  } /*  if (conspid > 0) */
  else {
    /* Code du fils consommateur */
    Consommateur(shmid);
  }
  /* QUESTION SUBSIDIAIRE : QUI (pre et/ou fils) VA EXECUTER CETTE */
  /* PARTIE DU CODE EN FIN D'EXECUTION DE(S) PROGRAMME(S) */
  pidmsg2("Processus","Fin du programme.");
  pidmsg3("Processus","le processus parent porte le numro (ppid=", getppid());
  return(EXIT_SUCCESS);
}
