/*----------------------------------------------------------------------*/
/* Fichier : tpi2ldd1simpletube.c                                       */
/* Auteur  : D. Duvivier - LIL/ULCO                                     */
/* Date    : 22/09/09                                                   */
/* Objectif: TP N3 M2/I2L Systme Rseaux Scurit                     */
/*           Test simple d'un tube et de la primitive "fork()" pour     */
/*           implementer une synchronisation de type                    */
/*           producteur-consommateur                                    */
/*           L'exemple est concu de maniere a generer un SIGPIPE        */
/*           intercepte (mais non traite completement) par le processus */
/*           pere/producteur suite a l'arret anticipe du fils.          */
/*----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include <signal.h>
#include <fcntl.h>
#include <unistd.h>
#include <limits.h> /* Pour taille ecriture atomique : PIPE_BUF... */

#define TAILLE_BUF      10 /* NE PAS MODIFIER : Taille du buffer utilise pour les E/S */
#define NB_ACCES        10 /* Nombre de lectures/ecritures */

/* Variables globales -> a viter/limiter normalement (:-(!!!) */
int i; /* indice de boucle */
int pid; /* pid du fils ou 0 */
char buffer[TAILLE_BUF+1];
int tube[2]; /* Un tube est un tableau de 2 descripteurs de fichiers */
int status_lecture, status_ecriture; /* Verif lecture/ecriture OK ? */
struct stat noeud_lecture, noeud_ecriture; /* Manipulation de structures */
int ncar; /* Nombre de caracteres effectivement lus/ecrits */

/* Gestion des signaux */
void signal_handler(int signal) {
  switch(signal) {
  case SIGHUP:
    /* Classiquement on force la relecture de la configuration */
    puts("Signal SIGHUP intercept par notre handler");
  break;
  case SIGPIPE:
    /* Classiquement on libre les ressources et on quitte */
    puts("Signal SIGPIPE intercept par notre handler");
    /* exit(99); */
  break;
  }
}


int main(void) {
  puts("Lancement du programme...");

  /* Le tube DOIT etre cree AVANT le fork afin que le pere et le fils */
  /* obtiennent les memes descripteurs de fichiers et puissent alors */
  /* communiquer par leur intermediaire */
  if (pipe(tube) < 0) {
    perror("echec creation tube");
    exit(EXIT_FAILURE);
  }
  /* Les appels suivants ne sont pas necessaires, ils montrent simplement */
  /* comment utiliser fcntl pour verifier le bon fonctionnement du tube */
  status_lecture=fcntl(tube[0], F_GETFL);
  if (status_lecture < 0) {
    perror("echec fcntl(lecture)");
    exit(EXIT_FAILURE);
  }
  status_ecriture=fcntl(tube[1], F_GETFL);
  if (status_ecriture < 0) {
    perror("echec fcntl(ecriture)");
    exit(EXIT_FAILURE);
  }
  printf("Attribut du descripteur %d tube[0] : %d\n", tube[0], status_lecture);
  printf("Attribut du descripteur %d tube[1] : %d\n", tube[1], status_ecriture);
  fflush(stdout);
  
  /* Avant de lancer le fork() et de partager le tube entre deux processus */
  /* je teste l'utilisation du tube depuis un seul processus comme une */
  /* simple file FIFO */
  /* Ecriture de TAILLE_BUF caracteres dans le tube */
  ncar = write(tube[1], "9876543210", TAILLE_BUF);
  if (ncar != TAILLE_BUF) {
    perror("echec lors de l'ecriture");
    exit(EXIT_FAILURE);
  }
  /* Lecture de TAILLE_BUF caracteres depuis le tube */
  ncar = read(tube[0], buffer, TAILLE_BUF);
  if (ncar != TAILLE_BUF) {
    perror("echec lors de la lecture");
    exit(EXIT_FAILURE);
  }
  buffer[ncar] = 0; /* Terminaison de chaine \0 par securite */
  printf("Chaine lue depuis le tube : %s (soit %d caracteres)\n",
         buffer, ncar);
  /* Ce qui suit n'est pas necessaire, les lignes suivantes montrent */
  /* simplement comment utiliser fstat et memcmp */
  /* Faites un "man fstat" pour plus d'informations ! */
  fstat(tube[0], &noeud_lecture);
  fstat(tube[1], &noeud_ecriture);
  if (memcmp(&noeud_lecture, &noeud_ecriture, sizeof(struct stat)) !=0) {
    perror("Noeuds en lecture et ecriture differents");
    exit(EXIT_FAILURE);    
  }
  puts("Noeuds en lecture et ecriture identiques");
  puts("Quelques caracteristiques du noeud :");
  if (S_ISFIFO(noeud_lecture.st_mode))
    puts("Type                      : tube");
  else
    puts("Type                      : autre que tube");
  printf("Numero de inode         : %d\n", (int)noeud_lecture.st_ino); 
  printf("Numero de peripherique  : %d\n", (int)noeud_lecture.st_dev); 
  printf("Nbre de liens physiques : %d\n", noeud_lecture.st_nlink); 
  printf("Taille                  : %d\n", (int)noeud_lecture.st_size); 
  printf("Nbre de blocs           : %d\n", (int)noeud_lecture.st_blocks); 
  printf("UID proprietaire        : %d\n", noeud_lecture.st_uid); 
  printf("GID proprietaire        : %d\n", noeud_lecture.st_gid); 
  printf("Protection/mode (octal) : %o\n", noeud_lecture.st_mode);
  printf("Ecriture par blocs de   : %d caracteres\n\n", PIPE_BUF);
  fflush(stdout);

  /*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/
  /* Il est possible de rendre la lecture depuis le tube non bloquante  */
  /* en utilisant la commande suivante :                                */
  /* fcntl(tube[0], F_SETFL, status_lecture | O_NONBLOCK);              */
  /* Mais en ce cas, il n'y aura plus de synchronisation de type        */
  /* producteur->consommateur puisque le consommateur ne se bloque plus */
  /* si le tube est vide !!!                                            */
  /*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/

  pid = fork();
  if (pid < 0) {
    perror("echec fork");
    exit(EXIT_FAILURE);
  }
  else if (pid > 0) {
    /* Code du pre */
    puts("Code du pre/producteur...");
    printf("Le pid du fils est %d\n", pid); fflush(stdout);

    /* Mise en place d'un gestionnaire de signaux pour le pere */
    signal(SIGHUP, signal_handler);
    signal(SIGPIPE, signal_handler);


    /* Le pere va uniquement ecrire dans le tube -> il FAUT fermer tube[0] */
    close(tube[0]);

    sleep(2);

    /* criture dans le tube */
    for (i=0; i<NB_ACCES; i++) {
      sleep(1);
      puts("Ecriture d'une chaine dans le tube..."); fflush(stdout);
      ncar = write(tube[1], "9876543210", TAILLE_BUF);
      if (ncar != TAILLE_BUF) {
        perror("echec lors de l'ecriture");
        exit(EXIT_FAILURE);
      }
    } 
    puts("Fin d'ecriture"); fflush(stdout);
    /* Fermer le descripteur encore ouvert */
    /* -> debloque le consommateur s'il est encore en attente */
    /* -> le nombre de caracteres lus peut etre = 0 en ce cas */
    close(tube[1]);

    puts("Fin du processus pre..."); fflush(stdout);
  }
  else {
    /* Code du fils */
    puts("Code du fils/consommateur...");
    printf("Le fils se met en attente du processus pere %d...\n", getppid());
    fflush(stdout);

    /* Le fils va uniquement lire dans le tube -> il FAUT fermer tube[1] */
    close(tube[1]);

    sleep(2);

    /* lecture depuis le tube */
    /* Ci-dessous j'ignore volontairement la derniere "lecture" */
    /* pour provoquer SIGPIPE */
    for (i=0; i<NB_ACCES-2; i++) {
      ncar = read(tube[0], buffer, TAILLE_BUF);
      if (ncar != TAILLE_BUF) {
        perror("echec lors de la lecture");
        exit(EXIT_FAILURE);
      }
      buffer[ncar] = 0; /* Terminaison de chaine \0 par securite */
      printf("Chaine lue depuis le tube : %s (soit %d caracteres)\n",
             buffer, ncar);
      fflush(stdout);
      sleep(1);
    }
    puts("Fin de lecture"); fflush(stdout);
    /* Fermer le descripteur encore ouvert */
    /* -> le producteur sera eventuellement averti par SIGPIPE */
    /* que le consommateur a casse/ferme le tube (broken pipe) */
    close(tube[0]);

    puts("Fin du processus fils..."); fflush(stdout);
  }
  /* Le pre et le fils vont passer ici en fin de programme sauf erreur */
  return(EXIT_SUCCESS);
}
